/* This code is public-domain - it is based on libcrypt
 * placed in the public domain by Wei Dai and other contributors.
 */
// gcc -Wall -DSHA1TEST -o sha1test sha1.c && ./sha1test

#include "sha1.h"

/* code */
#define SHA1_K0  0x5a827999
#define SHA1_K20 0x6ed9eba1
#define SHA1_K40 0x8f1bbcdc
#define SHA1_K60 0xca62c1d6

static uint32_t sha1_rol32(uint32_t number, uint8_t bits) {
    return ((number << bits) | (number >> (32-bits)));
}

static void sha1_hashBlock(sha1nfo *s) {
    uint8_t i;
    uint32_t a,b,c,d,e,t;

    a=s->state[0];
    b=s->state[1];
    c=s->state[2];
    d=s->state[3];
    e=s->state[4];
    for (i=0; i<80; i++) {
        if (i>=16) {
            t = s->buffer[(i+13)&15] ^ s->buffer[(i+8)&15] ^ s->buffer[(i+2)&15] ^ s->buffer[i&15];
            s->buffer[i&15] = sha1_rol32(t,1);
        }
        if (i<20) {
            t = (d ^ (b & (c ^ d))) + SHA1_K0;
        } else if (i<40) {
            t = (b ^ c ^ d) + SHA1_K20;
        } else if (i<60) {
            t = ((b & c) | (d & (b | c))) + SHA1_K40;
        } else {
            t = (b ^ c ^ d) + SHA1_K60;
        }
        t+=sha1_rol32(a,5) + e + s->buffer[i&15];
        e=d;
        d=c;
        c=sha1_rol32(b,30);
        b=a;
        a=t;
    }
    s->state[0] += a;
    s->state[1] += b;
    s->state[2] += c;
    s->state[3] += d;
    s->state[4] += e;
}

static void sha1_addUncounted(sha1nfo *s, uint8_t data) {
    uint8_t * const b = (uint8_t*) s->buffer;
#ifdef SHA_BIG_ENDIAN
    b[s->bufferOffset] = data;
#else
    b[s->bufferOffset ^ 3] = data;
#endif
    s->bufferOffset++;
    if (s->bufferOffset == BLOCK_LENGTH) {
        sha1_hashBlock(s);
        s->bufferOffset = 0;
    }
}

static void sha1_pad(sha1nfo *s) {
    // Implement SHA-1 padding (fips180-2 §5.1.1)

    // Pad with 0x80 followed by 0x00 until the end of the block
    sha1_addUncounted(s, 0x80);
    while (s->bufferOffset != 56) sha1_addUncounted(s, 0x00);

    // Append length in the last 8 bytes
    sha1_addUncounted(s, 0); // We're only using 32 bit lengths
    sha1_addUncounted(s, 0); // But SHA-1 supports 64 bit lengths
    sha1_addUncounted(s, 0); // So zero pad the top bits
    sha1_addUncounted(s, s->byteCount >> 29); // Shifting to multiply by 8
    sha1_addUncounted(s, s->byteCount >> 21); // as SHA-1 supports bitstreams as well as
    sha1_addUncounted(s, s->byteCount >> 13); // byte.
    sha1_addUncounted(s, s->byteCount >> 5);
    sha1_addUncounted(s, s->byteCount << 3);
}

void sha1_init(sha1nfo *s) {
    s->state[0] = 0x67452301;
    s->state[1] = 0xefcdab89;
    s->state[2] = 0x98badcfe;
    s->state[3] = 0x10325476;
    s->state[4] = 0xc3d2e1f0;
    s->byteCount = 0;
    s->bufferOffset = 0;
}

void sha1_writebyte(sha1nfo *s, uint8_t data) {
    ++s->byteCount;
    sha1_addUncounted(s, data);
}

void sha1_write(sha1nfo *s, const char *data, size_t len) {
    for (;len--;) sha1_writebyte(s, (uint8_t) *data++);
}

uint8_t* sha1_result(sha1nfo *s) {
    // Pad to complete the last block
    sha1_pad(s);

#ifndef SHA_BIG_ENDIAN
    // Swap byte order back
    int i;
    for (i=0; i<5; i++) {
        s->state[i] =
              (((s->state[i])<<24)& 0xff000000)
            | (((s->state[i])<<8) & 0x00ff0000)
            | (((s->state[i])>>8) & 0x0000ff00)
            | (((s->state[i])>>24)& 0x000000ff);
    }
#endif

    // Return pointer to hash (20 characters)
    return (uint8_t*) s->state;
}

/* self-test */

#ifdef SHA1TEST
#include <stdio.h>

void printHash(uint8_t* hash) {
    int i;
    for (i=0; i<20; i++) {
        printf("%02x", hash[i]);
    }
    printf("\n");
}


int main (int argc, char **argv) {
    uint32_t a;
    sha1nfo s;

    // SHA tests
    printf("Test: FIPS 180-2 C.1 and RFC3174 7.3 TEST1\n");
    printf("Expect:a9993e364706816aba3e25717850c26c9cd0d89d\n");
    printf("Result:");
    sha1_init(&s);
    sha1_write(&s, "abc", 3);
    printHash(sha1_result(&s));
    printf("\n\n");

    printf("Test: FIPS 180-2 C.2 and RFC3174 7.3 TEST2\n");
    printf("Expect:84983e441c3bd26ebaae4aa1f95129e5e54670f1\n");
    printf("Result:");
    sha1_init(&s);
    sha1_write(&s, "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", 56);
    printHash(sha1_result(&s));
    printf("\n\n");

    printf("Test: RFC3174 7.3 TEST4\n");
    printf("Expect:dea356a2cddd90c7a7ecedc5ebb563934f460452\n");
    printf("Result:");
    sha1_init(&s);
    for (a=0; a<80; a++) sha1_write(&s, "01234567", 8);
    printHash(sha1_result(&s));
    printf("\n\n");

    // Long tests
    printf("Test: FIPS 180-2 C.3 and RFC3174 7.3 TEST3\n");
    printf("Expect:34aa973cd4c4daa4f61eeb2bdbad27316534016f\n");
    printf("Result:");
    sha1_init(&s);
    for (a=0; a<1000000; a++) sha1_writebyte(&s, 'a');
    printHash(sha1_result(&s));

    return 0;
}
#endif /* self-test */
